"""
Ahorn
===================

A game description framework and game playing AI library,
written entirely in Python.

Quickstart
==========
    import ahorn, ahorn.Actors, ahorn.TicTacToe
    player_a, player_b = ahorn.Actors.MCTSPlayer(), ahorn.Actors.MCTSPlayer()
    starting_state = ahorn.TicTacToe.TicTacToeState([player_a, player_b])
    controller = ahorn.Controller(starting_state, verbose=True)
    controller.play()

Installation
============
    pip3 install -r requirements.txt
    python3 setup.py install
    python3 run.py  # should start playing a game

Running the tests
=================
    python3 -m pytest tests

Adding a new game
=================

A game is described by states and actions.
To describe a new game, subclass ahorn.GameBase.State and ahorn.GameBase.Action.
Take a look at the example: ahorn.TicTacToe.

Adding new AI
=============

Ahorn comes with a generic AI based on the Monte Carlo Tree Search algorithm:
 ahorn.Actors.MCTSPlayer.
To create a new AI, subclass ahorn.GameBase.Player. Take a look at the
example: ahorn.Actors.RandomPlayer.

"""
from .Controller import Controller



import abc

class Action(metaclass=abc.ABCMeta):
    """An action modifies one game state into the other.

    Parameters
    ----------

    Returns
    -------"""
    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def execute(self, state):
        """Perform the action on a given state

        Parameters
        ----------
        State:
            The state that must be modified

        Returns
        -------
        State:
            The modified state
        """
        pass

    @abc.abstractmethod
    def __str__(self):
        """A string representation of this action.

        Parameters
        ----------

        Returns
        -------
        str
            String representation of this action."""
        pass

    def __hash__(self):
        """Get a hash of the current action.

        Parameters
        ----------

        Returns
        -------
        int
            The hash of the action"""
        return hash(str(self))



"""Compare the strength of different Players"""
import random, statistics, time
import ahorn.Controller

class Arena(object):
    """In the arena you can test the strength of your Player on a given game

    Example:
        > player_a, player_b = MCTSPlayer(), MCTSPlayer()

        > arena = Arena(
            game=ahorn.TicTacToe.TicTacToeState,
            players=[player_a, player_b],
            n_games=20,
            verbose=True
        )

        > arena.play()

    Parameters
    ----------
    Game: State class
        The class of the game you want to play
    players: List<Player>
        The players you want to evaluate
    confidence: float
        The width of the confidence interval, default 90%
    verbose: bool
        Print some debug information
    verbose_seconds: float
        How many seconds between two prints, default 5

    Returns
    -------
    utilities: dict<player, float>
        The average utility for each player."""
    def __init__(self, Game, players, n_games=50, confidence=0.9, verbose=False, verbose_seconds=5):
        self.Game = Game
        self.players = players
        self.n_games = n_games
        self.confidence = confidence
        self.verbose = verbose
        self.verbose_seconds = verbose_seconds

    @staticmethod
    def bootstrap(series, func=statistics.mean, confidence=0.9):
        """Return the bootstrap confidence interval of a series.

        Parameters
        ----------
        series: List<float>
            your data
        func: function
            function that digests your data, default mean
        confidence: float
            width of the confidence interval, default 0.9

        Returns
        -------
        low: float
            lower bound of confidence interval
        mid: float
            median value of confidence interval
        high: float
            high bound of confidence interval"""
        n = len(series)
        n_bootstrap = 250
        digests = []
        for j in range(n_bootstrap):
            bootstrap_sample = [
                random.choice(series)
                for _ in range(n)
            ]
            digest = func(bootstrap_sample)
            digests.append(digest)
        digests.sort()
        low, mid, high = (1.0-confidence)/2.0, 0.5, (1.0+confidence)/2.0
        low, mid, high = int(low*n_bootstrap), int(mid*n_bootstrap), int(high*n_bootstrap)
        return digests[low], digests[mid], digests[high]

    def play(self):
        """Evaluate the strength of a player on a game.

        If verbose=True, will print intermediate results.

        Parameters
        ----------

        Returns
        -------
        result: dict<player, utility>
            the average utility of each player"""
        utilities = {
            player: []
            for player
            in self.players
        }
        start_time = time.time()
        prev_print = 0
        for j in range(self.n_games):
            random.shuffle(self.players)
            initial_state = self.Game(
                self.players
            )
            contr = ahorn.Controller(
                initial_state
            )
            final_state = contr.play()
            for player in self.players:
                utilities[player].append(final_state.get_utility(player))

            elapsed = time.time()-start_time
            elapsed_since_print = time.time()-prev_print
            if self.verbose and ((elapsed_since_print > self.verbose_seconds) or j == self.n_games-1):
                prev_print = time.time()
                print("{}".format(str(self.Game)))
                print(
                    "Game {} out of {} in {:2.1f}s ({:2.1f}s per game)".format(
                        j+1,
                        self.n_games,
                        elapsed,
                        elapsed/(j+1)
                    )
                )

                print("="*25)
                for player in sorted(self.players):
                    low, mid, high = Arena.bootstrap(
                        utilities[player],
                        func=statistics.mean,
                        confidence=self.confidence
                    )
                    print("{}\t|\t{:2.3f}/{:2.3f}/{:2.3f}".format(
                        str(player),
                        low,
                        mid,
                        high
                    ))
                print("")
        result = {
            player: Arena.bootstrap(
                utility,
                func=statistics.mean,
                confidence=self.confidence
            )[1]
            for player, utility
            in utilities.items()
        }
        return result



class Controller(object):
    """A controller is used to play a game

    The controller servers as the main interface to playing games.

    Parameters
    ----------
    players: List
        List of Player objects
    inititial_state: State
        State object from which to start the game

    Returns
    -------"""
    def __init__(self, initial_state, verbose=False):
        self.state = initial_state
        self.verbose = verbose

    def play(self):
        """Plays the game untill a final state is reached

        Parameters
        ----------

        Returns
        -------
        final_state: State
            The final state"""

        state = self.state
        while not state.is_final():
            actor = state.get_actor()
            if self.verbose: print(state.str(actor))
            action = actor.get_action(state.get_random(actor))
            assert(str(action) in [str(a) for a in state.get_legal_actions(actor)])
            state = action.execute(state)
            if self.verbose: print(str(action))
        self.state = state
        if self.verbose:
            print(state.str(actor))
            print("Points: {}".format(" ,".join([
                str(state.get_utility(player))
                for player in state.get_players()
            ])))

        return self.state



import abc

class State(metaclass=abc.ABCMeta):
    """Describes the attributes of a game at a particular time

    Parameters
    ----------

    Returns
    -------"""
    @abc.abstractmethod
    def __init__(self, players):
        self.players = players

    @abc.abstractmethod
    def copy(self, other):
        """Copy the content of another state into this state

        Deep copy, i.e. modifying the copied state can not influence the
        content of the original state.

        Parameters
        ----------
        other : State
            The state from which to copy the content

        Returns
        -------"""
        pass

    @abc.abstractmethod
    def is_final(self):
        """Return true if the state is final, i.e. the game is over

        Parameters
        ----------

        Returns
        -------
        bool
            True if the state is final, false otherwise"""
        pass

    @abc.abstractmethod
    def get_random(self, player):
        """Get a random sample from the information set of the current state.

        This method can be used to hide certain attributes of the state, such
        as opponent's cards. In case of games with perfect information this
        information contains only the state itself, and a copy of the state
        itself is returned.

        Parameters
        -----------
        player: Player
            The information set is created from the point of view of this actor

        Returns
        -------
        State
            A random sample from the information set"""
        pass

    @abc.abstractmethod
    def get_actor(self):
        """Return the actor that must perform an action in this state.

        Parameters
        ----------

        Returns
        -------
        Actor
            The actor that must perform an action in this state."""
        pass

    def get_players(self):
        """Return a list of all the players in the game.

        Parameters
        ----------

        Returns
        -------
        List<Player>
            A list of all the players in the game."""
        return self.players

    @abc.abstractmethod
    def get_legal_actions(self, actor):
        """Return the legal actions a player can take in this state.

        Parameters
        ----------
        actor: Actor
            the actor who wants to know which actions he can take

        Returns
        -------
        actions: List
            a list of Actions"""
        pass

    @abc.abstractmethod
    def get_utility(self, player):
        """Return the utility a given player receives.

        If the state is final, returns a utility, else returns None

        Parameters
        ----------
        player: Player
            The player for which to find the utility

        Returns
        -------
        utility: int
            The utility received by the player, or None"""
        pass

    @abc.abstractmethod
    def str(self, player):
        """String representation of this state, from the eyes of a given player.

        A string representation of the state, from the eyes of a given player.
        In games of incomplete information, this should hide the opponent cards,
        the deck, etc.
        In games of complete information, ignore the player.

        The string representation should be hashable, i.e. all states with the
        same hash should be in the same information set.

        Parameters
        ----------
        player: Player
            The player from who's eyes the state should be represented

        Returns
        -------
        str
            String representation of this state."""
        pass



from ..GameBase import State as BaseState
from .Actions import TicTacToeAction

class TicTacToeState(BaseState):
    """Describes a Tic-Tac-Toe state

    Parameters
    ----------
    players: List
        A list containing 2 Player objects

    Returns
    -------"""
    def __init__(self, players):
        self.board = [
            ["-", "-", "-"],
            ["-", "-", "-"],
            ["-", "-", "-"]
        ]
        assert(len(players) == 2)
        self.players = players
        self.pi = 0  # index of the current players

    def copy(self, other):
        """Copy the content of another state into this state

        Deep copy, i.e. modifying the copied state can not influence the
        content of the original state.

        Parameters
        ----------
        other : State
            The state from which to copy the content

        Returns
        -------"""

        for i, row in enumerate(other.board):
            for j, item in enumerate(row):
                self.board[i][j] = item
        self.pi = other.pi

    def _find_success(self):
        """Find OOO or XXX rows.

        Parameters
        ----------

        Returns
        -------
        row: List
            Either OOO or XXX, depending in which one has been found.
            Or None if none found"""
        def _row_identical(a, b):
            return all([
                aa == bb
                for aa, bb in zip(a, b)
            ])

        ooo = ["O", "O", "O"]
        xxx = ["X", "X", "X"]


        for row in self.board:  # check all the rows
            if _row_identical(row, ooo):
                return ooo
            if _row_identical(row, xxx):
                return xxx

        transpose = zip(*self.board)
        for collumn in transpose:  # check all the collumns
            if _row_identical(collumn, ooo):
                return ooo
            if _row_identical(collumn, xxx):
                return xxx

        diagonals = [
            [self.board[0][0], self.board[1][1], self.board[2][2]],
            [self.board[0][2], self.board[1][1], self.board[2][0]]
        ]
        for diagonal in diagonals:
            if _row_identical(diagonal, ooo):
                return ooo
            if _row_identical(diagonal, xxx):
                return xxx

        return None

    def is_final(self):
        """Return true if there is an OOO or XXX on the board, or the board is full.

        Parameters
        ----------

        Returns
        -------
        bool
            True if the state is an OXO on the board, false otherwise"""
        if self._find_success():  # OOO or XXX found, game is over
            return True

        free_spots = sum([
            1 if item == "-" else 0
            for row in self.board
            for item in row
        ])
        if free_spots == 0:  # no more free spots, game is over
            return True

        return False

    def get_random(self, player):
        """Tic-Tac-Toe is of complete information. Return a copy of this state.

        Parameters
        -----------
        player: Player
            Not used

        Returns
        -------
        State
            A copy of this state"""
        new = TicTacToeState(self.players)
        new.copy(self)
        return new

    def get_actor(self):
        """Return the actor that must perform an action in this state.

        Cycles between the first player and the second player

        Parameters
        ----------

        Returns
        -------
        Actor
            The player that must perform an action in this state."""
        return self.players[self.pi]

    def get_players(self):
        """Return a list of all the players in the game.

        Parameters
        ----------

        Returns
        -------
        List<Player>
            A list of all the players in the game."""
        return self.players

    def get_legal_actions(self, player):
        """Return the legal actions a player can take in this state.

        Parameters
        ----------
        player: Player
            the player who wants to know which actions he can take

        Returns
        -------
        actions: List
            a list of Actions"""
        symbol = ["O", "X"][self.players.index(player)]

        legal_actions = []
        for j, row in enumerate(self.board):
            for k, item in enumerate(row):
                if item == "-":
                    legal_actions.append(
                        TicTacToeAction(
                            symbol=symbol,
                            where=(j, k)
                        )
                    )
        return legal_actions

    def get_utility(self, player):
        """Return +1 if player won, -1 if player lost, 0 if draw.

        If the state is final, returns the utility, else returns None

        Parameters
        ----------
        player: Player
            The player for which to find the utility

        Returns
        -------
        utility: int
            The utility received by the player, or None"""
        if not self.is_final():
            return None

        success = self._find_success()
        if success:
            aim = ["O", "X"][self.players.index(player)]
            if success[0] == aim:
                return 1  # player achieved his goal and won
            else:
                return -1  # other player achieved his goal and won
        return 0

    def str(self, player):
        """Returns a string containing the board, and the current player

        Parameters
        ----------

        Returns
        -------
        str
            String representation of this state."""
        s = ""
        for row in self.board:
            s += "\t"+" ".join(row)+"\n"
        s += "Player {}'s turn".format(self.pi)
        return s

    def __str__(self):
        return "TicTacToe"



"""Tic-Tac-Toe is the simpelest of games.

The game is deterministic: there are no dice or random cards present.
The game is of complete information: every players knows the board.

This module serves as an example on how to implement simple games.
"""

from .Actions import TicTacToeAction
from .States import TicTacToeState



from ..GameBase import Action as BaseAction


class TicTacToeAction(BaseAction):
    """In Tac-Tac-Toe the only action is puttin an O or an X in a free space

    Parameters
    ----------
        symbol: str
            Either "X" or "O"
        where: (int, int)
            A tuple with row and column index

    Returns
    -------"""
    def __init__(self, symbol, where):
        self.symbol = symbol
        self.where = where

    def execute(self, state):
        """Execute the action.

        Modifies the board of the state, and the current player index.

        Parameters
        ----------
        state: TTTState
            The state which to modify

        Returns
        -------"""
        (r, c) = self.where
        state.board[r][c] = self.symbol
        state.pi = (state.pi+1) % 2
        return state

    def __str__(self):
        """A string representation of this action.

        For example: "Put a X in position 1, 2"

        Parameters
        ----------

        Returns
        -------
        str
            String representation of this action."""
        return "Put a {} in position {}".format(self.symbol, self.where)



"""
The Monte Carlo Tree Search player is a generic player, and can be used in games
of incomplete information and with chance events.

It is based on the Monte Carlo Tree Search algorithm.
"""
import random, time, math
from ..GameBase import Actor

class TreeNode(object):
    """A tree node to build internal game tree for MCTS algorithm

    Parameters
    ----------
    utility: List
        A utility vector with the utility for each player,
        for example [-1, 1] in a game with 2 players.
    simulations: int
        An number signifying how many times this node has been visited
        during the MCTS algorithm
    parent: TreeNode
        The parent of this node
    children: dict<Action -> TreeNode>
        A dictionary with actions as the keys, and treenodes as the values

    Returns
    -------"""
    def __init__(self, utility=None, simulations=0, parent=None, children=None):
        self.parent = parent
        self.simulations = simulations
        if not children:
            self.children = {}
        else:
            self.children = children

        if not utility:
            self.utility = []
        else:
            self.utility = utility

    def is_leaf():
        """True if this node is a leaf of the tree.

        Parameters
        ----------

        Returns
        -------
        is_leaf: Bool
            True if node is a leaf, False if node is not a leaf"""
        return len(self.children) == 0

class MCTSPlayer(Actor):
    """A player who uses the Monte Carlo Tree Search algorithm.

    Parameters
    ----------
    exploration_exploitation: float
        the exploration/exploitation parameter
    simulation_count: int
        maximum number of simulations
    simulation_time: float
        maximum seconds of simulation time
    verbose: bool
        print some information
    verboseverbose: bool
        print even more information

    Returns
    -------
    """
    def __init__(self, exploration_exploitation=1.4, simulation_count=10000, simulation_time=90, verbose=False, verboseverbose=False):
        self.verbose = verbose
        self.verboseverbose = verboseverbose
        self.exploration_exploitation = exploration_exploitation
        self.simulation_count = simulation_count
        self.simulation_time = simulation_time

    def get_action(self, state):
        """Return the best action according to MCTS algorithm

        Parameters
        ----------
        state: State
            The state in which the actor must perform an action

        Returns
        -------
        action: Action
            Best action according to MCTS algorithm"""
        if len(state.get_legal_actions(self)) == 1:
            return state.get_legal_actions(self)[0]
        n_players = len(state.get_players())
        # build a new tree
        root = TreeNode(
            utility=[0 for player in range(n_players)],
            simulations=0
        )
        start_time = time.time()
        simulations = 0

        while True:
            out_of_simulations = simulations >= self.simulation_count
            out_of_time = time.time() - start_time >= self.simulation_time
            if out_of_simulations or out_of_time:
                break

            random_state = state.get_random(self)
            selected_node, selected_state, action = self.selection_phase(root, random_state)
            expanded_node, expanded_state = self.expansion_phase(selected_node, selected_state, action)
            utility = self.simulation_phase(expanded_state)
            self.backpropagation_phase(expanded_node, utility)
            simulations += 1

        most_simulations, action_with_most_simulations = -1, None
        for action, tree_node in root.children.items():
            if tree_node.simulations > most_simulations:
                most_simulations = tree_node.simulations
                action_with_most_simulations = action

        if self.verbose:
            print("[MCTS] Total number of simulations: {} in {} seconds".format(str(simulations), time.time()-start_time))
            print("[MCTS] Possible action, average utility, simulations")
            for action, child in root.children.items():
                print(
                    "[MCTS] {}, {}, {}".format(
                        str(action),
                        str(["{0:0.2f}".format(u/child.simulations) for u in child.utility]),
                        str(child.simulations)
                    )
                )

        return action_with_most_simulations

    def selection_phase(self, root_node, root_state):
        """The selection phase of the MCTS algorithm.

        This phase will analyse the constructed MCTS tree,
        use UCT to traverse down the tree, until a root node is reached.

        Parameters
        ----------
        root_node: TreeNode
            The root tree node from where to start selection phase
        root_state: State
            The State corresponding to the root_node

        Returns
        -------
        selected_node: TreeNode
            The node in the MCTS tree that needs to be expanded
        selected_state: State
            The State corresponding to selected_node
        action: Action
            The action by which the selected node should be expanded with"""
        players = root_state.get_players()

        current_node = root_node
        current_state = root_state
        current_action = None

        while True:
            current_actor = current_state.get_actor()
            legal_actions = current_state.get_legal_actions(current_actor)
            # Check to see if there are actions in the current_state that have
            # not been simulated yet
            # if found, simulate them first
            for legal_action in legal_actions:
                children_str = [str(child) for child in current_node.children.keys()]
                if not str(legal_action) in children_str:
                    current_action = legal_action
                    return current_node, current_state, current_action

            is_random_state = not (current_actor in players)
            if is_random_state:
                # If this is a random state (for example taking a card from deck)
                # Return one of the actions randomly
                current_action = random.choice(legal_actions)
            else:
                # This is a player state, use UCT rule to select the action
                current_player_index = players.index(current_actor)
                ntotal = current_node.simulations

                best_uct, best_action = -float('inf'), None
                for action, child_state in current_node.children.items():
                    ni = child_state.simulations
                    win_percentage = child_state.utility[current_player_index]/ni
                    confidence = math.sqrt(math.log(ntotal)/ni)
                    uct = win_percentage + self.exploration_exploitation*confidence
                    if uct > best_uct:
                        best_uct = uct
                        best_action = action

                current_action = best_action

            current_node = current_node.children[current_action]
            current_state = current_action.execute(current_state)
            if current_state.is_final():
                # premature end because the constructed MCTS tree
                # contains a path from root to leaf in the real game tree
                return current_node, current_state, current_action

    def expansion_phase(self, selected_node, selected_state, action):
        """The expansion phase of the MCTS algorithm.

        This expansion will add a node to the constructed MCTS tree,

        Parameters
        ----------
        selected_node: TreeNode
            The node which needs to be expanded
        selected_state: State
            The State corresponding to the selected_node
        action: Action
            The action by which the selected node needs to be expanded

        Returns
        -------
        expanded_node: TreeNode
            The new node added to the constructed MCTS tree
        expanded_state: State
            The State corresponding to expanded_node"""
        expanded_node = TreeNode(
            parent=selected_node,
            utility=[0 for _ in selected_state.get_players()],
        )
        selected_node.children[action] = expanded_node
        expanded_state = action.execute(selected_state)
        return expanded_node, expanded_state

    def simulation_phase(self, expanded_state):
        """The simulation phase of the MCTS algorithm.

        This phase will simulate a random playout from a given state

        Parameters
        ----------
        expanded_state: State
            The state from where to start the random playout

        Returns
        -------
        utility: List
            The utility vector of the final state"""
        state = expanded_state
        while not state.is_final():
            actor = state.get_actor()
            actions = state.get_legal_actions(actor)
            action = random.choice(actions)
            if self.verboseverbose:
                print("[MCTS][simulation]: {}".format(str(action)))
            state = action.execute(state)
            if self.verboseverbose:
                print(state.str(actor))
        return [
            state.get_utility(player)
            for player in state.get_players()
        ]

    def backpropagation_phase(self, expanded_node, utility):
        """The backpropagation phase of the MCTS algorithm.

        This phase will update the constructed MCTS tree based on the
        utility vector from the simulation

        Parameters
        ----------
        expanded_node: TreeNode
            The node from where to start backpropagation
        utility: List<float>
            The utility vector from the simulation phase

        Returns
        -------"""
        node = expanded_node
        while node:
            node.simulations += 1
            node.utility = [
                old + new
                for old, new
                in zip(node.utility, utility)
            ]
            node = node.parent

    def __str__(self):
        return "MCTSPlayer"



import abc

class Actor(metaclass=abc.ABCMeta):
    """An actor performs actions and drives a game from state to state.

    Parameters
    ----------

    Returns
    -------
    """
    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def get_action(self, state):
        """Return the action the actor wants to take in a given state.

        Parameters
        ----------
        State:
            The state in which the actor must perform an action

        Returns
        -------
        action: Action
            The action the actor wants to take in this state."""
        pass

    @abc.abstractmethod
    def __str__(self):
        """A name for this actor

        Parameters
        ----------

        Returns
        -------
        name: str
            The name for this actor."""
        pass



from . import RandomActor
from ..GameBase import Player

class RandomPlayer(RandomActor, Player):
    """A player who takes a random action from the list of legal actions."""
    pass
    def __str__(self):
        return "RandomPlayer"



from distutils.core import setup

setup(
  name = 'ahorn',
  packages = ['ahorn', 'ahorn.Actors', 'ahorn.GameBase', 'ahorn.TicTacToe'],
  version = '0.3',
  description = 'A game playing and game AI library',
  author = 'Dries Wijns',
  author_email = 'dries.wijns@gmail.com',
  url = 'https://github.com/drieswijns/ahorn',
  download_url = 'https://github.com/drieswijns/ahorn/tarball/master',
  keywords = ['game', 'playing', 'AI', 'MCTS'],
  classifiers = [],
  package_data={"ahorn": ["ahorn/*"]},
)



import random
from ..GameBase import Actor


class RandomActor(Actor):
    """An actor who always choses random actions.

    Parameters
    ----------

    Returns
    -------"""
    def __init__(self):
        pass

    def get_action(self, state):
        """Return a random action from the legal actions

        Parameters
        ----------
        State:
            The state in which the actor must perform an action

        Returns
        -------
        Action:
            A random action from the legal actions"""
        return random.choice(
            state.get_legal_actions(self)
        )

    def __str__(self):
        return "RandomActor"
